"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || (function () {
    var ownKeys = function(o) {
        ownKeys = Object.getOwnPropertyNames || function (o) {
            var ar = [];
            for (var k in o) if (Object.prototype.hasOwnProperty.call(o, k)) ar[ar.length] = k;
            return ar;
        };
        return ownKeys(o);
    };
    return function (mod) {
        if (mod && mod.__esModule) return mod;
        var result = {};
        if (mod != null) for (var k = ownKeys(mod), i = 0; i < k.length; i++) if (k[i] !== "default") __createBinding(result, mod, k[i]);
        __setModuleDefault(result, mod);
        return result;
    };
})();
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.authDataProvider = void 0;
exports.purgeTokenCache = purgeTokenCache;
exports.authorizeWithUserPass = authorizeWithUserPass;
exports.getModuleConfig = getModuleConfig;
exports.authorizeWithRefreshToken = authorizeWithRefreshToken;
exports.isTokenExpired = isTokenExpired;
exports.createAuthDataProvider = createAuthDataProvider;
exports.setAuthData = setAuthData;
exports.clearAuthData = clearAuthData;
exports.getLoggedUserData = getLoggedUserData;
exports.updateUserData = updateUserData;
exports.deleteUser = deleteUser;
const axios_1 = __importStar(require("axios"));
const apiTools_1 = require("./apiTools");
const mutex_1 = require("./utils/mutex");
const jwt_decode_1 = require("jwt-decode");
const my_events_1 = require("my-events");
const isEqual_1 = __importDefault(require("lodash/isEqual"));
const _authDataMap = new Map();
const _moduleConfigMap = new Map();
const mutex = new mutex_1.Mutex(250);
const invalidRefreshTokens = new Map();
class TokenExpirationEmitter extends my_events_1.EventEmitter {
    constructor() {
        super();
    }
    notify(tenant, error) {
        this.emit(`refreshTokenInvalid-${tenant}`, error);
    }
}
const RefreshTokenNotifier = new TokenExpirationEmitter();
function purgeTokenCache() {
    const keys = invalidRefreshTokens.keys();
    for (const key of keys) {
        invalidRefreshTokens.delete(key);
    }
}
function authorizeWithUserPass(baseUrl, tenant, basicAuthPass, username, password) {
    return __awaiter(this, void 0, void 0, function* () {
        let response = null;
        try {
            response = yield axios_1.default.post(`${baseUrl}/auth-oauth2/oauth/token`, { username, password, grant_type: 'password', scope: 'read' }, {
                headers: {
                    'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8',
                    Authorization: `Basic ${basicAuthPass}`,
                    'X-Tenant': tenant,
                },
            });
            _authDataMap.set(tenant, response.data);
            return true;
        }
        catch (error) {
            //console.error('Authorization error');
            //console.error(error);
            return false;
        }
    });
}
/**
 * Gets module config only if given accessToken is for a module, otherwise returns undefined
 * @param baseUrl
 * @param accessToken
 */
function getModuleConfig(baseUrl, accessToken) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            const tokenData = (0, jwt_decode_1.jwtDecode)(accessToken);
            if (tokenData.MODULE) {
                const { data } = yield axios_1.default.get(`${baseUrl}/auth-api/api/module-config`, {
                    headers: {
                        Authorization: `Bearer ${accessToken}`,
                    },
                });
                return data;
            }
        }
        catch (e) {
            console.warn('get module config failed', e);
            return undefined;
        }
    });
}
function authorizeWithRefreshToken(baseUrl, tenant, basicAuthPass, refreshToken, moduleConfigCallback, fingerprint) {
    return __awaiter(this, void 0, void 0, function* () {
        var _a, _b, _c;
        let response = null;
        try {
            response = yield axios_1.default.post(`${baseUrl}/auth-oauth2/oauth/token`, {
                refresh_token: refreshToken,
                grant_type: 'refresh_token',
                fingerprint: fingerprint || undefined,
            }, {
                headers: {
                    'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8',
                    Authorization: `Basic ${basicAuthPass}`,
                    'X-Tenant': tenant,
                },
            });
            _authDataMap.set(tenant, response.data);
            // if module get & refresh module config
            void getModuleConfig(baseUrl, response.data.access_token).then((moduleConfig) => {
                if (!moduleConfig)
                    return;
                const id = `${tenant}-${moduleConfig.id}`;
                const current = _moduleConfigMap.get(id);
                _moduleConfigMap.set(id, moduleConfig);
                if (!current || !(0, isEqual_1.default)(current, moduleConfig)) {
                    // update via callback if has changed
                    console.log('module config changed prev', current, 'new', moduleConfig);
                    moduleConfigCallback === null || moduleConfigCallback === void 0 ? void 0 : moduleConfigCallback(moduleConfig);
                }
            });
            // console.log('authData', _authData);
            //console.log('REFRESH TOKEN AUTH RESULT ***');
            //console.log(_authData);
            return { success: true };
        }
        catch (error) {
            if ((0, axios_1.isAxiosError)(error) && /^4\d{2}$/.test(`${(_a = error.response) === null || _a === void 0 ? void 0 : _a.status}`)) {
                invalidRefreshTokens.set(refreshToken, Date.now());
                if (((_c = (_b = error.response) === null || _b === void 0 ? void 0 : _b.data) === null || _c === void 0 ? void 0 : _c.error) === 'invalid_grant') {
                    console.warn('invalid_grant', tenant);
                    RefreshTokenNotifier.notify(tenant, error);
                    // token invalid or expired, remove it
                    return { success: false, removeToken: true };
                }
            }
            //console.error('Authorization error');
            // console.error(error);
            return { success: false, removeToken: false };
        }
    });
}
function isTokenExpired(token) {
    // check if we can use current token
    if (token) {
        try {
            return (
            // 10s before actual expiry
            (0, jwt_decode_1.jwtDecode)(token).exp * 1000 < new Date().getTime() + 10000);
        }
        catch (error) {
            // Not a JWT token
            return true;
        }
    }
    return true;
}
function isJwtToken(token) {
    try {
        (0, jwt_decode_1.jwtDecode)(token);
        return true;
    }
    catch (error) {
        return false;
    }
}
const authDataProvider = (ctx_1, refreshTokenStorageHandler_1, ...args_1) => __awaiter(void 0, [ctx_1, refreshTokenStorageHandler_1, ...args_1], void 0, function* (ctx, refreshTokenStorageHandler, forceRefresh = false, moduleConfigCallback, fingerprint = null) {
    // console.log( `authDataProvider invoked ----     currentToken=${  _authData ? _authData.access_token : '(null)'   }`, );
    let authData = _authDataMap.get(ctx.TENANT);
    if (!forceRefresh && authData && authData.access_token) {
        // console.log('auth data has access token');
        if (!isTokenExpired(authData.access_token)) {
            return { token: authData.access_token, UUID: authData.UUID };
        }
        else {
            console.log('access token is expired');
        }
    }
    //console.log('******** GET AUTH TOKEN (REMOTE CALL) ********');
    let loginSuccess = false;
    const refreshToken = yield refreshTokenStorageHandler.getRefreshToken(ctx.TENANT);
    console.log('have refresh token?', !!refreshToken, 'tenant', ctx.TENANT);
    if (isJwtToken(refreshToken) && isTokenExpired(refreshToken)) {
        refreshTokenStorageHandler.clearRefreshToken(ctx.TENANT);
    }
    if (refreshToken &&
        (!isJwtToken(refreshToken) || !isTokenExpired(refreshToken))) {
        //console.log('*** AUTHORIZE WITH REFRESH TOKEN ');
        // use mutex to enforce max 1 api call at a time
        const result = yield mutex.getLock(() => {
            authData = _authDataMap.get(ctx.TENANT);
            if (authData && !isTokenExpired(authData.access_token)) {
                return Promise.resolve({ success: true });
            }
            // If same refresh token was invalid less then 30s ago skip api call
            const invalidTokenTimestamp = invalidRefreshTokens.get(refreshToken);
            if (invalidTokenTimestamp && Date.now() < invalidTokenTimestamp + 30000) {
                return Promise.resolve({ success: false, removeToken: false });
            }
            return authorizeWithRefreshToken(ctx.BASE_URL, ctx.TENANT, ctx.BASIC_AUTH, refreshToken, moduleConfigCallback, fingerprint);
        });
        loginSuccess = result.success;
        if (result.removeToken) {
            yield refreshTokenStorageHandler.clearRefreshToken(ctx.TENANT);
        }
    }
    if (!loginSuccess && ctx.anonymousAuth) {
        // console.log('**** ANONYMOUS AUTH ****');
        loginSuccess = yield authorizeWithUserPass(ctx.BASE_URL, ctx.TENANT, ctx.BASIC_AUTH, 'anonymous', '');
    }
    authData = _authDataMap.get(ctx.TENANT);
    if (!loginSuccess || !authData) {
        return { token: '', UUID: '' };
    }
    else {
        refreshTokenStorageHandler.setRefreshToken(ctx.TENANT, authData.refresh_token);
        return { token: authData.access_token, UUID: authData.UUID };
    }
});
exports.authDataProvider = authDataProvider;
function createAuthDataProvider(ctx, refreshTokenProvider, forceRefresh = false, refreshTokenErrorCallback, moduleConfigCallback, fingerprint) {
    RefreshTokenNotifier.on(`refreshTokenInvalid-${ctx.TENANT}`, (error) => {
        refreshTokenErrorCallback === null || refreshTokenErrorCallback === void 0 ? void 0 : refreshTokenErrorCallback(error);
    });
    return () => (0, exports.authDataProvider)(ctx, refreshTokenProvider, forceRefresh, moduleConfigCallback, fingerprint);
}
function setAuthData(tenant, newAuthData, refreshTokenStorageHandler) {
    _authDataMap.set(tenant, newAuthData);
    refreshTokenStorageHandler.setRefreshToken(tenant, newAuthData.refresh_token);
}
function clearAuthData(tenant, refreshTokenStorageHandler) {
    refreshTokenStorageHandler.clearRefreshToken(tenant);
    _authDataMap.delete(tenant);
}
// fetch user data from /auth-api/api/me
function getLoggedUserData(baseURL, token) {
    return __awaiter(this, void 0, void 0, function* () {
        const response = yield axios_1.default
            .get(`${baseURL}/auth-api/api/me`, {
            headers: {
                'Content-Type': 'application/json',
                Authorization: `Bearer ${token}`,
            },
        })
            .catch(apiTools_1.handleAPIError);
        return response ? response.data : null;
    });
}
function updateUserData(baseURL, token, userData) {
    return __awaiter(this, void 0, void 0, function* () {
        const response = yield axios_1.default
            .post(`${baseURL}/auth-api/api/me`, userData, {
            headers: {
                'Content-Type': 'application/json',
                Authorization: `Bearer ${token}`,
            },
        })
            .catch(apiTools_1.handleAPIError);
        return (response && response.data) || false;
    });
}
function deleteUser(baseURL, token) {
    return __awaiter(this, void 0, void 0, function* () {
        const data = yield getLoggedUserData(baseURL, token);
        if (!data) {
            return false;
        }
        const { login } = data;
        const response = yield axios_1.default
            .request({
            url: `${baseURL}/auth-api/api/me`,
            method: 'DELETE',
            headers: {
                'Content-Type': 'application/json',
                Authorization: `Bearer ${token}`,
            },
            data: { login },
        })
            .catch(apiTools_1.handleAPIError);
        return (response && response.data) || false;
    });
}
